@@ -1,4 +1,4 @@
-/* $Id: curl.c,v 1.35 2001/12/06 11:44:12 kjetilja Exp $ */
+/* $Id: curl.c,v 1.52 2002/03/08 14:12:39 kjetilja Exp $ */
 
 /* cURL Python module by Kjetil Jacobsen <kjetilja @ cs.uit.no> */
 
@@ -9,42 +9,56 @@
 #include <stdio.h>
 #include <assert.h>
 
+/* Ensure we have an updated libcurl */
+#if LIBCURL_VERSION_NUM < 0x070905
+  #error "Need libcurl version 7.9.5 or greater to compile pycurl."
+#endif
+
 static PyObject *ErrorObject;
 
 typedef struct {
     PyObject_HEAD
     CURL *handle;
-    char *url;
     struct HttpPost *httppost;
     struct curl_slist *httpheader;
     struct curl_slist *quote;
     struct curl_slist *postquote;
+    struct curl_slist *prequote;
     PyObject *w_cb;
     PyObject *h_cb;
     PyObject *r_cb;
     PyObject *pro_cb;
     PyObject *pwd_cb;
     PyThreadState *state;
+    int writeheader_set;
     char error[CURL_ERROR_SIZE];
+    void *options[CURLOPT_LASTENTRY];
 } CurlObject;
 
 #if !defined(__cplusplus)
 staticforward PyTypeObject Curl_Type;
 #endif
 
+#define CURLERROR() \
+{\
+    PyObject *v; \
+    v = Py_BuildValue("(is)", res, self->error); \
+    PyErr_SetObject(ErrorObject, v); \
+    Py_DECREF(v); \
+    return NULL; \
+}
+
 /* --------------------------------------------------------------------- */
 
 static void
 self_cleanup(CurlObject *self)
 {
+    int i;
+
     if (self->handle != NULL) {
 	curl_easy_cleanup(self->handle);
         self->handle = NULL;
     }
-    if (self->url != NULL) {
-	free(self->url);
-	self->url = NULL;
-    }
     if (self->httpheader != NULL) {
 	curl_slist_free_all(self->httpheader);
 	self->httpheader = NULL;
@@ -57,10 +71,21 @@ self_cleanup(CurlObject *self)
 	curl_slist_free_all(self->postquote);
 	self->postquote = NULL;
     }
+    if (self->prequote != NULL) {
+	curl_slist_free_all(self->prequote);
+	self->prequote = NULL;
+    }
     if (self->httppost != NULL) {
 	curl_formfree(self->httppost);
 	self->httppost = NULL;
     }
+    for (i = 0; i < CURLOPT_LASTENTRY; i++) {
+	if (self->options[i] != NULL) {
+	    free(self->options[i]);
+	    self->options[i] = NULL;
+	}		
+    }	
+    self->state = NULL;
     Py_XDECREF(self->w_cb);
     Py_XDECREF(self->r_cb);
     Py_XDECREF(self->pro_cb);
@@ -109,6 +134,11 @@ write_callback(void *ptr,
     self = (CurlObject *)stream;
     arglist = Py_BuildValue("(s#)", (char *)ptr, size*nmemb);
 
+    /* Check whether we got a file object or a curl object */
+    if (self->state == NULL) {
+	return -1;
+    }
+
     PyEval_AcquireThread(self->state);
     result = PyEval_CallObject(self->w_cb, arglist);
     Py_DECREF(arglist);
@@ -116,10 +146,12 @@ write_callback(void *ptr,
 	PyErr_Print();
 	write_size = -1;
     }
-    else if (result == Py_None)               /* None means success */
+    else if (result == Py_None) {               /* None means success */
         write_size = (int)(size * nmemb);
-    else
+    }
+    else {
         write_size = (int)PyInt_AsLong(result);
+    }
     Py_XDECREF(result);
     PyEval_ReleaseThread(self->state);
     return write_size;
@@ -140,6 +172,11 @@ header_callback(void *ptr,
     self = (CurlObject *)stream;
     arglist = Py_BuildValue("(s#)", (char *)ptr, size*nmemb);
 
+    /* Check whether we got a file object or a curl object */
+    if (self->state == NULL) {
+	return -1;
+    }
+
     PyEval_AcquireThread(self->state);
     result = PyEval_CallObject(self->h_cb, arglist);
     Py_DECREF(arglist);
@@ -147,10 +184,12 @@ header_callback(void *ptr,
 	PyErr_Print();
 	write_size = -1;
     }
-    else if (result == Py_None)               /* None means success */
+    else if (result == Py_None) {               /* None means success */
         write_size = (int)(size * nmemb);
-    else
+    }
+    else {
         write_size = (int)PyInt_AsLong(result);
+    }
     Py_XDECREF(result);
     PyEval_ReleaseThread(self->state);
     return write_size;
@@ -172,6 +211,11 @@ progress_callback(void *client,
     self = (CurlObject *)client;
     arglist = Py_BuildValue("(iiii)", dltotal, dlnow, ultotal, ulnow);
 
+    /* Check whether we got a file object or a curl object */
+    if (self->state == NULL) {
+	return -1;
+    }
+
     PyEval_AcquireThread(self->state);
     result = PyEval_CallObject(self->pro_cb, arglist);
     Py_DECREF(arglist);
@@ -179,10 +223,12 @@ progress_callback(void *client,
 	PyErr_Print();
 	ret = -1;
     }
-    else if (result == Py_None)               /* None means success */
+    else if (result == Py_None) {               /* None means success */
         ret = 0;
-    else
+    }
+    else {
 	ret = (int)PyInt_AsLong(result);
+    }
     Py_XDECREF(result);
     PyEval_ReleaseThread(self->state);
     return ret;
@@ -204,6 +250,11 @@ int password_callback(void *client,
     self = (CurlObject *)client;
     arglist = Py_BuildValue("(si)", prompt, buflen);
 
+    /* Check whether we got a file object or a curl object */
+    if (self->state == NULL) {
+	return -1;
+    }
+
     PyEval_AcquireThread(self->state);
     result = PyEval_CallObject(self->pwd_cb, arglist);
     Py_DECREF(arglist);
@@ -253,6 +304,11 @@ int read_callback(void *ptr,
     read_size = size*nmemb;
     arglist = Py_BuildValue("(i)", read_size);
 
+    /* Check whether we got a file object or a curl object */
+    if (self->state == NULL) {
+	return -1;
+    }
+
     PyEval_AcquireThread(self->state);
     result = PyEval_CallObject(self->r_cb, arglist);
     Py_DECREF(arglist);
@@ -295,7 +351,7 @@ int read_callback(void *ptr,
 static PyObject *
 do_setopt(CurlObject *self, PyObject *args)
 {
-    int option;
+    int option, opt_masked;
     char *stringdata;
     long longdata;
     char *buf;
@@ -337,35 +393,38 @@ do_setopt(CurlObject *self, PyObject *args)
 	      option == CURLOPT_RANDOM_FILE ||
 	      option == CURLOPT_COOKIEJAR ||
 	      option == CURLOPT_SSL_CIPHER_LIST ||
-	      option == CURLOPT_EGDSOCKET))
+	      option == CURLOPT_EGDSOCKET ||
+	      option == CURLOPT_SSLCERTTYPE ||
+	      option == CURLOPT_SSLKEY ||
+	      option == CURLOPT_SSLKEYTYPE ||
+	      option == CURLOPT_SSLKEYPASSWD ||
+	      option == CURLOPT_SSLENGINE))
 	    {
 		PyErr_SetString(ErrorObject, "strings are not supported for this option");
 		return NULL;
 	    }
-	if (option == CURLOPT_URL) {
-	    /* Need to store uri for later use if the option is OPTCURL_URL */
-	    buf = (char *)malloc((strlen(stringdata)*sizeof(char))+sizeof(char));
-	    if (buf == NULL)
-	        return PyErr_NoMemory();
-	    if (self->url != NULL) {
-		free(self->url);
-	    }
-	    strcpy(buf, stringdata);
-	    self->url = buf;
-	    res = curl_easy_setopt(self->handle, CURLOPT_URL, self->url);
+	/* Free previously allocated memory to option */
+	opt_masked = option % CURLOPTTYPE_OBJECTPOINT;
+	if (self->options[opt_masked] != NULL) {
+	    free(self->options[opt_masked]);
 	}
-	else {
-	    /* Handle the regular cases of string arguments */
-	    res = curl_easy_setopt(self->handle, option, stringdata);
+	/* Allocate memory to hold the string */
+	buf = (char *)malloc((strlen(stringdata)*sizeof(char))+sizeof(char));
+	if (buf == NULL) {
+	    return PyErr_NoMemory();
 	}
+	strcpy(buf, stringdata);
+	self->options[opt_masked] = buf;
+	/* Call setopt */
+	res = curl_easy_setopt(self->handle, option, 
+			       (char *)self->options[opt_masked]);
 	/* Check for errors */
 	if (res == 0) {
 	    Py_INCREF(Py_None);
 	    return Py_None;
 	}
 	else {
-	    PyErr_SetString(ErrorObject, self->error);
-	    return NULL;
+	    CURLERROR();
 	}
     }
 
@@ -385,8 +444,7 @@ do_setopt(CurlObject *self, PyObject *args)
 	    return Py_None;
 	}
 	else {
-	    PyErr_SetString(ErrorObject, self->error);
-	    return NULL;
+	    CURLERROR();
 	}
     }
 
@@ -404,6 +462,13 @@ do_setopt(CurlObject *self, PyObject *args)
 		PyErr_SetString(PyExc_TypeError, "files are not supported for this option");
 		return NULL;
 	    }
+        if (option == CURLOPT_WRITEHEADER) {
+	    self->writeheader_set = 1;
+            if (self->w_cb != NULL) {
+                PyErr_SetString(ErrorObject, "cannot combine WRITEHEADER with WRITEFUNCTION.");
+		return NULL;
+            }
+        }
 	fp = PyFile_AsFile(obj);
 	if (fp == NULL) {
 	    PyErr_SetString(PyExc_TypeError, "second argument must be open file");
@@ -416,8 +481,7 @@ do_setopt(CurlObject *self, PyObject *args)
 	    return Py_None;
 	}
 	else {
-	    PyErr_SetString(ErrorObject, self->error);
-	    return NULL;
+  	    CURLERROR();
 	}
     }
 
@@ -435,6 +499,9 @@ do_setopt(CurlObject *self, PyObject *args)
 	case CURLOPT_POSTQUOTE:
 	    slist = &self->postquote;
 	    break;
+	case CURLOPT_PREQUOTE:
+	    slist = &self->prequote;
+	    break;
 	case CURLOPT_HTTPPOST:
 	    slist = NULL;
 	    break;
@@ -467,8 +534,7 @@ do_setopt(CurlObject *self, PyObject *args)
 		res = curl_formparse(buf, &self->httppost, &last);
 		if (res != 0) {
 		    curl_formfree(self->httppost);
-		    PyErr_SetString(ErrorObject, self->error);
-		    return NULL;
+                    CURLERROR();
 		}
 	    }
 	    res = curl_easy_setopt(self->handle, CURLOPT_HTTPPOST, self->httppost);
@@ -479,8 +545,7 @@ do_setopt(CurlObject *self, PyObject *args)
 	    }
 	    else {
 		curl_formfree(self->httppost);
-		PyErr_SetString(ErrorObject, self->error);
-		return NULL;
+		CURLERROR();
 	    }
 	}
 
@@ -516,15 +581,13 @@ do_setopt(CurlObject *self, PyObject *args)
 	}
 	else {
 	    curl_slist_free_all(*slist);
-	    PyErr_SetString(ErrorObject, self->error);
-	    return NULL;
+	    CURLERROR();
 	}
     }
 
     PyErr_Clear();
 
     /* Handle the case of function objects for callbacks */
-
     if (PyArg_ParseTuple(args, "iO!:setopt", &option, &PyFunction_Type, &obj) ||
 	PyArg_ParseTuple(args, "iO!:setopt", &option, &PyCFunction_Type, &obj) ||
 	PyArg_ParseTuple(args, "iO!:setopt", &option, &PyMethod_Type, &obj))
@@ -533,6 +596,10 @@ do_setopt(CurlObject *self, PyObject *args)
 
 	switch(option) {
 	case CURLOPT_WRITEFUNCTION:
+	    if (self->writeheader_set == 1) {
+	        PyErr_SetString(ErrorObject, "cannot combine WRITEFUNCTION with WRITEHEADER option.");
+	        return NULL;
+	    }
 	    Py_INCREF(obj);
 	    Py_XDECREF(self->w_cb);
 	    self->w_cb = obj;
@@ -589,6 +656,7 @@ do_perform(CurlObject *self, PyObject *args)
 {
     int res;
 
+    /* Sanity checks */
     if (!PyArg_ParseTuple(args, ":perform")) {
 	return NULL;
     }
@@ -599,7 +667,7 @@ do_perform(CurlObject *self, PyObject *args)
     }
 
     /* Save handle to current thread (used to run the callbacks in) */
-    self->state = PyThreadState_New(PyThreadState_Get()->interp);
+    self->state = PyThreadState_Get();
 
     /* Release global lock and start */
     Py_BEGIN_ALLOW_THREADS
@@ -611,8 +679,7 @@ do_perform(CurlObject *self, PyObject *args)
 	return Py_None;
     }
     else {
-	PyErr_SetString(ErrorObject, self->error);
-	return NULL;
+        CURLERROR();
     }
 }
 
@@ -647,12 +714,12 @@ do_getinfo(CurlObject *self, PyObject *args)
 		    return PyLong_FromLong(l_res);
 		}
 		else {
-		    PyErr_SetString(ErrorObject, self->error);
-		    return NULL;
+		    CURLERROR();
 		}
 	    }
 
-	if (option == CURLINFO_EFFECTIVE_URL)
+	if (option == CURLINFO_EFFECTIVE_URL ||
+            option == CURLINFO_CONTENT_TYPE)
 	    {
 		/* Return string as result */
 		res = curl_easy_getinfo(self->handle, option, &s_res);
@@ -661,8 +728,7 @@ do_getinfo(CurlObject *self, PyObject *args)
 		    return PyString_FromString(s_res);
 		}
 		else {
-		    PyErr_SetString(ErrorObject, self->error);
-		    return NULL;
+		    CURLERROR();
 		}
 	    }
 
@@ -685,8 +751,7 @@ do_getinfo(CurlObject *self, PyObject *args)
 		    return PyFloat_FromDouble(d_res);
 		}
 		else {
-		    PyErr_SetString(ErrorObject, self->error);
-		    return NULL;
+		    CURLERROR();
 		}
 	    }
     }
@@ -701,7 +766,7 @@ do_getinfo(CurlObject *self, PyObject *args)
 static char co_cleanup_doc [] = "cleanup() -> None.  End curl session.\n";
 static char co_setopt_doc [] = "setopt(option, parameter) -> None.  Set curl session options.  Throws pycurl.error exception upon failure.\n";
 static char co_perform_doc [] = "perform() -> None.  Perform a file transfer.  Throws pycurl.error exception upon failure.\n";
-static char co_getinfo_doc [] = "getinfo(info, parameter) -> None.  Extract information from a curl session.  Throws pycurl.error upon failure.\n";
+static char co_getinfo_doc [] = "getinfo(info, parameter) -> res.  Extract and return information from a curl session.  Throws pycurl.error exception upon failure.\n";
 
 
 static PyMethodDef curlobject_methods[] = {
@@ -761,11 +826,12 @@ do_init(PyObject *arg)
 
     /* Setup python curl object initial values */
     self->handle = NULL;
-    self->url = NULL;
     self->httpheader = NULL;
     self->quote = NULL;
     self->postquote = NULL;
+    self->prequote = NULL;
     self->httppost = NULL;
+    self->writeheader_set = 0;
 
     /* Set callback pointers to NULL */
     self->w_cb = NULL;
@@ -784,12 +850,16 @@ do_init(PyObject *arg)
     if (res != 0)
         goto error;
     memset(self->error, 0, sizeof(char) * CURL_ERROR_SIZE);
+    
+    /* Zero memory buffer for setopt */
+    memset(self->options, 0, sizeof(void *) * CURLOPT_LASTENTRY);
 
-    /* Set NOPROGRESS to 1 by default */
+    /* Enable NOPROGRESS by default */
     res = curl_easy_setopt(self->handle, CURLOPT_NOPROGRESS, 1);
     if (res != 0)
         goto error;
-    /* Set VERBOSE to 0 by default */
+
+    /* Disable VERBOSE by default */
     res = curl_easy_setopt(self->handle, CURLOPT_VERBOSE, 0);
     if (res != 0)
         goto error;
@@ -949,6 +1019,7 @@ DL_EXPORT(void)
     insint(d, "CRLF", CURLOPT_CRLF);
     insint(d, "QUOTE", CURLOPT_QUOTE);
     insint(d, "POSTQUOTE", CURLOPT_POSTQUOTE);
+    insint(d, "PREQUOTE", CURLOPT_PREQUOTE);
     insint(d, "WRITEHEADER", CURLOPT_WRITEHEADER);
     insint(d, "HEADERFUNCTION", CURLOPT_HEADERFUNCTION);
     insint(d, "COOKIEFILE", CURLOPT_COOKIEFILE);
@@ -984,6 +1055,13 @@ DL_EXPORT(void)
     insint(d, "HTTP_VERSION_1_1", CURL_HTTP_VERSION_1_1);
     insint(d, "FTP_USE_EPSV", CURLOPT_FTP_USE_EPSV);
 
+    insint(d, "SSLCERTTYPE", CURLOPT_SSLCERTTYPE); 
+    insint(d, "SSLKEY", CURLOPT_SSLKEY);
+    insint(d, "SSLKEYTYPE", CURLOPT_SSLKEYTYPE);
+    insint(d, "SSLKEYPASSWD", CURLOPT_SSLKEYPASSWD);
+    insint(d, "SSLENGINE", CURLOPT_SSLENGINE);
+    insint(d, "SSLENGINE_DEFAULT", CURLOPT_SSLENGINE_DEFAULT);
+
     /* Symbolic constants for getinfo */
     insint(d, "EFFECTIVE_URL", CURLINFO_EFFECTIVE_URL);
     insint(d, "HTTP_CODE", CURLINFO_HTTP_CODE);
@@ -1002,6 +1080,7 @@ DL_EXPORT(void)
     insint(d, "CONTENT_LENGTH_DOWNLOAD", CURLINFO_CONTENT_LENGTH_DOWNLOAD);
     insint(d, "CONTENT_LENGTH_UPLOAD", CURLINFO_CONTENT_LENGTH_UPLOAD);
     insint(d, "STARTTRANSFER_TIME", CURLINFO_STARTTRANSFER_TIME);
+    insint(d, "CONTENT_TYPE", CURLINFO_CONTENT_TYPE);
 
     /* CLOSEPOLICY constants for setopt */
     insint(d, "CLOSEPOLICY_LEAST_RECENTLY_USED", CURLCLOSEPOLICY_LEAST_RECENTLY_USED);
@@ -1022,4 +1101,3 @@ DL_EXPORT(void)
 /*
 vi:ts=8
 */
-
